/*----------------------------------------------------------------------------
Multiboot-compatible test kernel
Chris Giese <geezer@execpc.com> http://my.execpc.com/~geezer

I, the copyright holder of this work, hereby release it into the
public domain. This applies worldwide. If this is not legally possible:
I grant any entity the right to use this work for any purpose,
without any conditions, unless such conditions are required by law.

Feb. 25, 2014:
- now displays 256-color .BMP files in VGA mode 0x13 (320x200) instead of
  16-color .PCX files in VGA mode 0x12 (640x480)
- memory map (from INT 15h AX=E820h) was not being displayed properly; fixed
- VGA register dump for text mode is now read from the VGA itself;
  not hard-coded
- added init_video() function in VIDEO.C to get actual screen size
  from the BIOS data area
- got rid of the "inc" and "lib" directories and put all source code
  into two files
- my newest printf() code
- "boot_dev" changed to "root_dev"

Nov. 14, 2003:
- initial release
----------------------------------------------------------------------------*/
/*****************************************************************************
INC/STDINT.H
*****************************************************************************/
typedef signed char		int8_t;
typedef short			int16_t;
typedef long			int32_t;
#if defined(__GNUC__)
typedef long long		int64_t;
#endif

typedef unsigned char		uint8_t;
typedef unsigned short		uint16_t;
typedef unsigned long		uint32_t;
#if defined(__GNUC__)
typedef unsigned long long	unt64_t;
#endif
/*****************************************************************************
KRNL/MLTIBOOT.H
*****************************************************************************/
/* bits in mboot_info_t->flags */
#define MBINFO_MEM		0x00000001
#define MBINFO_ROOT_DEV		0x00000002
#define MBINFO_CMD_LINE		0x00000004
#define MBINFO_MODS		0x00000008
//#define MBINFO_AOUT_SYMS	0x00000010
//#define MBINFO_ELF_SHDR	0x00000020
#define MBINFO_MEM_MAP		0x00000040
#define MBINFO_DRIVE_INFO	0x00000080
#define MBINFO_CONFIG_TABLE	0x00000100
#define MBINFO_LOADER_NAME	0x00000200
#define MBINFO_APM_TABLE	0x00000400
#define MBINFO_VIDEO_INFO	0x00000800

#pragma pack(1)
typedef struct
{
	uint32_t flags;
	uint32_t conv_mem;	/* MBINFO_MEM */
	uint32_t ext_mem;	/* MBINFO_MEM */
	uint32_t root_dev;	/* MBINFO_ROOT_DEV */
	uint32_t cmd_line;	/* MBINFO_CMD_LINE */
/* modules */
	uint32_t num_mods;	/* MBINFO_MODS */
	uint32_t mods_addr;	/* MBINFO_MODS */
/* no symbol table (MBINFO_AOUT_SYMS) or section table (MBINFO_ELF_SHDR) */
	uint32_t unused3[4];
/* BIOS memory map */
	uint32_t map_len;	/* MBINFO_MEM_MAP */
	uint32_t map_addr;	/* MBINFO_MEM_MAP */
/* info about hard drives */
	uint32_t drives_len;	/* MBINFO_DRIVE_INFO */
	uint32_t drives_addr;	/* MBINFO_DRIVE_INFO */
/* ROM config table */
	uint32_t config_table;	/* MBINFO_CONFIG_TABLE */
/* boot loader name */
	uint32_t loader_name;	/* MBINFO_LOADER_NAME */
/* APM table. APM is busted in GRUB 0.97 */
	uint32_t apm_table;	/* MBINFO_APM_TABLE */
/* video info -- optional in GRUB 0.97, I think */
	uint32_t vbe_ctrl_info;	/* MBINFO_VIDEO_INFO */
	uint32_t vbe_mode_info;	/* MBINFO_VIDEO_INFO */
	uint16_t vbe_mode;	/* MBINFO_VIDEO_INFO */
	uint16_t vbe_iface_seg;	/* MBINFO_VIDEO_INFO */
	uint16_t vbe_iface_off;	/* MBINFO_VIDEO_INFO */
	uint16_t vbe_iface_len;	/* MBINFO_VIDEO_INFO */
} mboot_info_t;

/* Multiboot modules */
typedef struct
{
	uint32_t start_addr;
	uint32_t end_addr;
	uint32_t cmd_line;
	uint32_t unused;
} mboot_mod_t;

/* Multiboot BIOS memory ranges */
typedef struct
{
	uint32_t size; /* =20 =(size of this struct - 4) */
	uint32_t addr, res_addr;
	uint32_t len, res_len;
	uint32_t type;
} mboot_range_t;

/* I _think_ this struct should be packed...not sure */
#pragma pack(1)
typedef struct
{
	uint32_t len;
	uint8_t int13_drive_num;
	uint8_t use_lba;
	uint16_t cyls;
	uint8_t heads;
	uint8_t sects;
/* Adjust 'len' if this array is longer than 1. GRUB 0.97 actually
doesn't do anything with this field but I think there needs to be
at least one zero value here... */
	uint16_t io_ports[1];
} mboot_drive_t;
/*----------------------------------------------------------------------------
LIBRARY ROUTINES
----------------------------------------------------------------------------*/
/*****************************************************************************
INC/HARDWARE.H
*****************************************************************************/
void outportb(unsigned port, unsigned val);
unsigned char inportb(unsigned port);
/*****************************************************************************
LIB/HARDWARE/OUTPORTB.C
*****************************************************************************/
void outportb(unsigned port, unsigned val)
{
	__asm__(
		"outb %b0,%w1"
		:
		: "a"(val), "d"(port));
}
/*****************************************************************************
LIB/HARDWARE/INPORTB.C
*****************************************************************************/
unsigned char inportb(unsigned port)
{
	unsigned ret_val;

	__asm__ ("inb %w1,%b0"
		: "=a"(ret_val)
		: "d"(port));
	return ret_val;
}
/*****************************************************************************
INC/STRING.H
*****************************************************************************/
#define	NULL	0

typedef unsigned size_t;

size_t strlen(const char *str);
void *memcpy(void *dst_ptr, const void *src_ptr, size_t count);
/*****************************************************************************
LIB/STRING/STRLEN.C
*****************************************************************************
#include <string.h> size_t */

size_t strlen(const char *str)
{
	size_t ret_val;

	for(ret_val = 0; *str != '\0'; str++)
		ret_val++;
	return ret_val;
}
/*****************************************************************************
LIB/STRING/MEMCPY.C
******************************************************************************
#include <string.h> size_t */

void *memcpy(void *dst_ptr, const void *src_ptr, size_t count)
{
	const char *src = src_ptr;
	char *dst = dst_ptr;
	void *ret_val = dst_ptr;

	for(; count != 0; count--)
	{
		*dst = *src;
		dst++;
		src++;
	}
	return ret_val;
}
/*****************************************************************************
INC/STDARG.H

EXPORTS:
va_start()
va_arg()
va_end()
va_list

Warning: this code assumes standard _Cdecl calling convention; as used by
Turbo C (16-bit x86) and GCC for 32-bit x86. 64-bit x86 and Watcom C use
a register calling convention.
*****************************************************************************/
/* Assume: width of stack == width of int. Don't use sizeof(char *) or
other pointer because sizeof(char *)==4 for LARGE-model 16-bit code.
Assume: width is a power of 2 */
#define	STACK_WIDTH	sizeof(int)

/* Round up object width so it's an even multiple of STACK_WIDTH.
Using AND for modulus here, so STACK_WIDTH must be a power of 2. */
#define	TYPE_WIDTH(TYPE)				\
	((sizeof(TYPE) + STACK_WIDTH - 1) & ~(STACK_WIDTH - 1))

/* point the va_list pointer to LASTARG,
then advance beyond it to the first variable arg */
#define	va_start(PTR, LASTARG)				\
	PTR = (va_list)((char *)&(LASTARG) + TYPE_WIDTH(LASTARG))

/* Increment the va_list pointer, then "return"
(evaluate to, actually) the previous value of the pointer. */
#define va_arg(PTR, TYPE)	(			\
	PTR = (char *)(PTR) + TYPE_WIDTH(TYPE)		\
				,			\
	*((TYPE *)((char *)(PTR) - TYPE_WIDTH(TYPE)))	\
				)

#define va_end(PTR)	/* nothing */

typedef void *va_list;
/*****************************************************************************
INC/STDIO.H
*****************************************************************************/
#define	EOF		(-1)

void printf(const char *fmt, ...);
/*****************************************************************************
INC/_PRINTF.H
******************************************************************************
#include <stdarg.h> va_list */

typedef int (*fnptr_t)(unsigned c, void **helper);

int do_printf(const char *fmt, va_list args, fnptr_t fn, void *ptr);
/*****************************************************************************
LIB/STDIO/DOPRINTF.C

EXPORTS:
int do_printf(const char *fmt, va_list args, fnptr_t fn, void *ptr);

xxx - current implementation of printf("%n" ...) is wrong -- RTFM

Revised Feb. 16, 2014:
- test for NULL buffer moved from sprintf() to vsprintf()

Revised Dec. 3, 2013
- do_printf() restructured to get rid of confusing goto statements
- do_printf() now returns EOF if an error occurs
  (currently, this happens only if out-of-memory in vasprintf_help())
- added vasprintf() and asprintf()
- added support for %Fs (far pointer to string)
- compile-time option (PRINTF_UCP) to display pointer values
  as upper-case hex (A-F), instead of lower-case (a-f)
- the code to check for "%--6d", "%---6d", etc. has been removed;
  these are now treated the same as "%-6d"

Revised Feb. 3, 2008
- sprintf() now works with NULL buffer; returns size of output
- changed va_start() macro to make it compatible with ANSI C

Revised Dec. 12, 2003
- fixed vsprintf() and sprintf() in test code

Revised Jan. 28, 2002
- changes to make characters 0x80-0xFF display properly

Revised June 10, 2001
- changes to make vsprintf() terminate string with '\0'

Revised May 12, 2000
- math in DO_NUM (currently num2asc()) is now unsigned, as it should be
- %0 flag (pad left with zeroes) now works
- actually did some TESTING, maybe fixed some other bugs
******************************************************************************
#include <stdarg.h> va_list, va_arg()
#include <_printf.h> fnptr_t */

/* display pointers in upper-case hex (A-F) instead of lower-case (a-f) */
#define	PRINTF_UCP	1

/* flags used in processing format string */
#define	PR_LEFTJUST	0x01	/* left justify */
#define	PR_UC		0x02	/* use upper case (A-F vs. a-f) for hex */
#define	PR_DO_SIGN	0x04	/* signed numeric conversion (%d vs. %u) */
#define	PR_LONG		0x08	/* long numeric conversion (%lu) */
#define	PR_SHORT	0x10	/* short numeric conversion (%hu) */
#define	PR_NEGATIVE	0x20	/* PR_DO_SIGN set and num was < 0 */
#define	PR_PADLEFT0	0x40	/* pad left with '0' instead of ' ' */
#define	PR_FARPTR	0x80	/* pointers are far */
/*****************************************************************************
Converts binary numeric value 'num' to ASCII string.
'buf' must point to a zeroed byte at the END of a suitably-large buffer.
Return value is a pointer to the START of the string stored in 'buf'.
*****************************************************************************/
static unsigned char *num2asc(unsigned char *buf, unsigned radix,
		unsigned long num, int upper_case)
{
	unsigned long i;

/* do..while() instead of while() ensures we get a single '0' if num==0 */
	do
	{
		i = (unsigned long)num % radix;
		buf--;
		if(i < 10)
			*buf = i + '0';
		else
		{
			if(upper_case)
				*buf = i - 10 + 'A';
			else
				*buf = i - 10 + 'a';
		}
		num = (unsigned long)num / radix;
	} while(num != 0);
	return buf;
}
/*****************************************************************************
name:	do_printf
action:	minimal subfunction for ?printf, calls function
	'fn' with arg 'ptr' for each character to be output
returns:total number of characters output
*****************************************************************************/
int do_printf(const char *fmt, va_list args, fnptr_t fn, void *ptr)
{
	unsigned flags, actual_wd, given_wd, count;
/* Largest number handled is 2^32-1, lowest radix handled is 8.
2^32-1 in base 8 has 11 digits (add 5 for trailing NUL and for slop).
Strings are copied directly to the output, and don't use 'buf'. */
#define	PR_BUFLEN	16
	unsigned char state, radix, buf[PR_BUFLEN];
	unsigned char *where;
	long num;
	int i;

	state = count = given_wd = flags = 0;
/* for() and switch() on the same line looks jarring
but the indentation gets out of hand otherwise */
	for(; *fmt; fmt++) switch(state)
	{
/* STATE 0: AWAITING '%' */
	case 0:
/* echo text until '%' seen */
		if(*fmt != '%')
		{
			if((i = fn(*fmt, &ptr)) == EOF)
				return i;
			count++;
			break;
		}
/* found %, get next char and advance state to check if next char is a flag */
		fmt++;
		state++;
		flags = 0;
		/* FALL THROUGH */
/* STATE 1: AWAITING FLAGS ('%' or '-' or '0') */
	case 1:
		if(*fmt == '%')	/* %% */
		{
			if((i = fn(*fmt, &ptr)) == EOF)
				return i;
			count++;
			state = 0;
			break;
		}
		if(*fmt == '-')
		{
			flags |= PR_LEFTJUST;
			break;
		}
		if(*fmt == '0')
		{
			flags |= PR_PADLEFT0;
/* '0' could be flag or field width -- fall through */
			fmt++;
		}
/* '0' or not a flag char: advance state to check if it's field width */
		state++;
		given_wd = 0;
		/* FALL THROUGH */
/* STATE 2: AWAITING (NUMERIC) FIELD WIDTH */
	case 2:
		if(*fmt >= '0' && *fmt <= '9')
		{
			given_wd = 10 * given_wd + (*fmt - '0');
			break;
		}
/* not field width: advance state to check if it's a modifier */
		state++;
		/* FALL THROUGH */
/* STATE 3: AWAITING MODIFIER CHARACTERS ('F', 'N', 'l', or 'h') */
	case 3:
		if(*fmt == 'F')
		{
			flags |= PR_FARPTR;
			break;
		}
		if(*fmt == 'N') /* near pointer (the default) */
		{
			flags &= ~PR_FARPTR;
			break;
		}
		if(*fmt == 'l')
		{
			flags |= PR_LONG;
			break;
		}
		if(*fmt == 'h')
		{
			flags |= PR_SHORT;
			break;
		}
/* not a modifier: advance state to check if it's a conversion char */
		state++;
		/* FALL THROUGH */
/* STATE 4: AWAITING CONVERSION CHARACTER
('X' 'x' 'p' 'n' 'd' 'i' 'u' 'o' 'c' 's') */
	case 4:
		where = &buf[PR_BUFLEN - 1];
		*where = '\0';
/* 'h' and 'l' modifiers cancel each other out */
		if((flags & (PR_LONG | PR_SHORT)) == (PR_LONG | PR_SHORT))
			flags &= ~(PR_LONG | PR_SHORT);
/* pointer and numeric conversions */
		if(*fmt == 'p' || *fmt == 'n'
			|| *fmt == 'X' || *fmt == 'x'
			|| *fmt == 'd' || *fmt == 'i'
			|| *fmt == 'u' || *fmt == 'o')
		{
#if defined(PRINTF_UCP)
			if(*fmt == 'X' || *fmt == 'p' || *fmt == 'n')
#else
			if(*fmt == 'X')
#endif
			{
				flags |= PR_UC;
				flags &= ~PR_DO_SIGN;
				radix = 16; /* display pointers in hex */
			}
#if defined(PRINTF_UCP)
			else if(*fmt == 'x')
#else
			else if(*fmt == 'x' || *fmt == 'p' || *fmt == 'n')
#endif
			{
				flags &= ~PR_UC;
				flags &= ~PR_DO_SIGN;
				radix = 16;
			}
			else if(*fmt == 'd' || *fmt == 'i')
			{
				flags |= PR_DO_SIGN;
				radix = 10;
			}
			else if(*fmt == 'u')
			{
				flags &= ~PR_DO_SIGN;
				radix = 10;
			}
			else /*if(*fmt == 'o')*/
			{
				flags &= ~PR_DO_SIGN;
				radix = 8;
			}
/* load the value to be printed. 'l'=long (usu. 32 bits)... */
			if(flags & PR_LONG)
			{
				if(flags & PR_DO_SIGN)
					num = va_arg(args, signed long);
				else
					num = va_arg(args, unsigned long);
			}
/* ...'h'=short (usu. 16 bits) */
			else if(flags & PR_SHORT)
			{
				if(flags & PR_DO_SIGN)
					num = va_arg(args, signed short);
				else
					num = va_arg(args, unsigned short);
			}
/* no 'h' nor 'l': sizeof(int) bits */
			else
			{
				if(flags & PR_DO_SIGN)
					num = va_arg(args, signed int);
				else
					num = va_arg(args, unsigned int);
			}
/* convert to unsigned but remember that it was negative */
			if((flags & PR_DO_SIGN) && num < 0)
			{
				flags |= PR_NEGATIVE;
				num = -num;
			}
/* convert binary unsigned number to octal/decimal/hex ASCII */
			where = num2asc(where, radix, num, flags & PR_UC);
		}
		else if(*fmt == 'c')
		{
/* disallow these modifiers for %c */
			flags &= ~(PR_DO_SIGN | PR_NEGATIVE | PR_PADLEFT0);
/* yes; we're converting a character to a string here: */
			where--;
			*where = (unsigned char)va_arg(args, unsigned char);
		}
		else if(*fmt == 's')
		{
/* disallow these modifiers for %s */
			flags &= ~(PR_DO_SIGN | PR_NEGATIVE | PR_PADLEFT0);
			where = va_arg(args, unsigned char *);
		}
/* bogus conversion character -- copy it to output and go back to state 0 */
		else
		{
			if((i = fn(*fmt, &ptr)) == EOF)
				return i;
			count++;
			state = flags = given_wd = 0;
			break;
		}
/* emit formatted string */
		actual_wd = strlen(where);
		if(flags & PR_NEGATIVE)
			actual_wd++;
/* if we pad left with ZEROES, do the sign now
(for numeric values; not for %c or %s) */
		if((flags & (PR_NEGATIVE | PR_PADLEFT0))
			== (PR_NEGATIVE | PR_PADLEFT0))
		{
			if((i = fn('-', &ptr)) == EOF)
				return i;
			count++;
		}
/* pad on left with spaces or zeroes (for right justify) */
		if((flags & PR_LEFTJUST) == 0)
		{
			for(; given_wd > actual_wd; given_wd--)
			{
				if((i = fn(flags & PR_PADLEFT0
					? '0' : ' ', &ptr)) == EOF)
						return i;
				count++;
			}
/* if we pad left with SPACES, do the sign now */
			if(flags & PR_NEGATIVE)
			{
				if((i = fn('-', &ptr)) == EOF)
					return i;
				count++;
			}
		}
/* emit converted number/char/string */
		for(; *where != '\0'; where++)
		{
			if((i = fn(*where, &ptr)) == EOF)
				return i;
			count++;
		}
/* pad on right with spaces (for left justify) */
		if(given_wd < actual_wd)
			given_wd = 0;
		else
			given_wd -= actual_wd;
		for(; given_wd; given_wd--)
		{
			if((i = fn(' ', &ptr)) == EOF)
				return i;
			count++;
		}
		/* FALL THROUGH */
	default:
		state = flags = given_wd = 0;
		break;
	}
	return count;
}
/*****************************************************************************
LIB/STDIO/PRINTF.C
******************************************************************************
#include <stdarg.h> va_list, va_start(), va_end()
#include <_printf.h> do_printf() */

/* IMPORTS
from KRNL/CONSOLE.C */
int putch(unsigned c);

static int printf_help(unsigned c, void **ptr_UNUSED)
{
/* normal printf() uses putchar()
	return putchar(c); */
	return putch(c);
}
/****************************************************************************/
void printf(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	do_printf(fmt, args, printf_help, NULL);
	va_end(args);
}
/*----------------------------------------------------------------------------
KRNL/CONSOLE.C

EXPORTS:
int putch(unsigned c);
void clear_screen(void);
void init_video(void);
unsigned getch(void);
------------------------------------------------------------------------------
#include <hardware.h> outportb(), inportb()
#include <stdint.h> uint8_t, uint16_t */

static uint16_t *g_fb_addr;
static const unsigned g_attrib = 0x1E; /* blue background, yellow text */
static unsigned g_csr_x, g_csr_y, g_vc_wd, g_vc_ht, g_crtc_io;
/****************************************************************************/
static void move_csr(void)
{
	unsigned i;

	i = g_vc_wd * g_csr_y + g_csr_x;
	outportb(g_crtc_io + 0, 14);
	outportb(g_crtc_io + 1, i >> 8);
	outportb(g_crtc_io + 0, 15);
	outportb(g_crtc_io + 1, i);
}
/****************************************************************************/
static void memsetw(void *dst_ptr, int val, size_t count)
{
	uint16_t *dst = dst_ptr;

	for(; count != 0; count--)
	{
		*dst = val;
		dst++;
	}
}
/****************************************************************************/
int putch(unsigned c)
{
/* backspace */
	if(c == 0x08)
	{
		if(g_csr_x != 0)
			g_csr_x--;
	}
/* tab */
	else if(c == 0x09)
		g_csr_x = (g_csr_x + 8) & ~7;
/* carriage return */
	else if(c == '\r')	/* 0x0D */
		g_csr_x = 0;
/* line feed */
//	else if(c == '\n')	/* 0x0A */
//		g_csr_y++;
/* CR/LF */
	else if(c == '\n')
	{
		g_csr_x = 0;
		g_csr_y++;
	}
/* printable ASCII */
	else if(c >= ' ')
	{
		uint16_t *where;

		where = &g_fb_addr[g_vc_wd * g_csr_y + g_csr_x];
		*where = (g_attrib << 8) | c;
		g_csr_x++;
	}
	if(g_csr_x >= g_vc_wd)
	{
		g_csr_x = 0;
		g_csr_y++;
	}
/* scroll up */
	if(g_csr_y >= g_vc_ht)
	{
		unsigned i;

		i = g_csr_y - g_vc_ht + 1;
		memcpy(g_fb_addr, &g_fb_addr[i * g_vc_wd],
			(g_vc_ht - i) * g_vc_wd * 2);
/* erase the bottom line of the screen */
		memsetw(&g_fb_addr[(g_vc_ht - i) * g_vc_wd],
			(g_attrib << 8) | (unsigned)' ', g_vc_wd);
		g_csr_y = g_vc_ht - i;
	}
	move_csr();
	return 0;
}
/****************************************************************************/
void clear_screen(void)
{
	memsetw(g_fb_addr, (g_attrib << 8) | (unsigned)' ',
		g_vc_ht * g_vc_wd);
	g_csr_x = g_csr_y = 0;
}
/****************************************************************************/
void init_video(void)
{
/* read text-mode screen size from the BIOS data area */
	g_vc_wd = *(uint16_t *)0x44A;
	g_vc_ht = *(uint8_t *)0x484 + 1;
#define	VGA_MISC_READ	0x3CC
	if((inportb(VGA_MISC_READ) & 0x01) == 0)
	{
		g_fb_addr = (uint16_t *)0xB0000L;
		g_crtc_io = 0x3B4;	/* monochrome emulation */
	}
	else
	{
		g_fb_addr = (uint16_t *)0xB8000L;
		g_crtc_io = 0x3D4;	/* color emulation */
	}
	clear_screen();
}
/****************************************************************************/
unsigned getch(void)
{
	unsigned gray = 0, rv;

	while(1)
	{
/* await byte from 8042 keyboard controller chip */
		while((inportb(0x64) & 0x01) == 0)
			/* nothing */;
/* get byte */
		rv = inportb(0x60);
/* prefix byte for "gray" keys (arrow keys, numeric keypad, etc.) */
		if(rv == 0xE0)
		{
			gray = 0x100;
			continue;
		}
/* make code */
		if(rv < 0x80)
			break;
/* break code */
		gray = 0;
	}
	return gray | rv;
}
/*----------------------------------------------------------------------------
KRNL/VGA.C

EXPORTS:
void set_graphics_mode(void);
void set_text_mode(void);
------------------------------------------------------------------------------
#include <hardware.h> outportb(), inportb()
#include <stdint.h> uint16_t, uint8_t
#include <string.h> memcpy() */

/* IMPORTS
from CONSOLE.C */
unsigned getch(void);

/* Attribute Controller */
#define	VGA_AC_INDEX		0x3C0
#define	VGA_AC_WRITE		0x3C0
#define	VGA_AC_READ		0x3C1
/* Miscellaneous register */
#define	VGA_MISC_WRITE		0x3C2
/* Sequencer */
#define VGA_SEQ_INDEX		0x3C4
#define VGA_SEQ_DATA		0x3C5
/* DAC (i.e. the 256-color palette) */
#define	VGA_DAC_READ_INDEX	0x3C7
#define	VGA_DAC_WRITE_INDEX	0x3C8
#define	VGA_DAC_DATA		0x3C9
#define	VGA_MISC_READ		0x3CC
/* Graphics Controller */
#define VGA_GC_INDEX 		0x3CE
#define VGA_GC_DATA 		0x3CF
/* CRT Controller	COLOR emulation		MONO emulation */
#define VGA_CRTC_INDEX		0x3D4		/* 0x3B4 */
#define VGA_CRTC_DATA		0x3D5		/* 0x3B5 */
#define	VGA_INSTAT_READ		0x3DA

#define	VGA_NUM_SEQ_BYTES	5
#define	VGA_NUM_CRTC_BYTES	25
#define	VGA_NUM_GC_BYTES	9
#define	VGA_NUM_AC_BYTES	21
#define	VGA_NUM_DAC_BYTES	768
#define	VGA_NUM_BYTES		(1 + VGA_NUM_SEQ_BYTES + VGA_NUM_CRTC_BYTES \
				+ VGA_NUM_GC_BYTES + VGA_NUM_AC_BYTES \
				+ VGA_NUM_DAC_BYTES)

static char g_text_init;
static unsigned char g_font[4096];
static unsigned char g_text[VGA_NUM_BYTES];
/****************************************************************************/
static void write_regs(unsigned char *regs)
{
	unsigned i;

/* write MISCELLANEOUS reg */
	outportb(VGA_MISC_WRITE, *regs);
	regs++;
/* write SEQUENCER regs */
	for(i = 0; i < VGA_NUM_SEQ_BYTES; i++)
	{
		outportb(VGA_SEQ_INDEX, i);
		outportb(VGA_SEQ_DATA, *regs);
		regs++;
	}
/* unlock CRTC registers */
	outportb(VGA_CRTC_INDEX, 0x03);
	outportb(VGA_CRTC_DATA, inportb(VGA_CRTC_DATA) | 0x80);
	outportb(VGA_CRTC_INDEX, 0x11);
	outportb(VGA_CRTC_DATA, inportb(VGA_CRTC_DATA) & ~0x80);
/* make sure stored CRTC values don't re-lock the registers */
	regs[0x03] |= 0x80;
	regs[0x11] &= ~0x80;
/* write CRTC regs */
	for(i = 0; i < VGA_NUM_CRTC_BYTES; i++)
	{
		outportb(VGA_CRTC_INDEX, i);
		outportb(VGA_CRTC_DATA, *regs);
		regs++;
	}
/* write GRAPHICS CONTROLLER regs */
	for(i = 0; i < VGA_NUM_GC_BYTES; i++)
	{
		outportb(VGA_GC_INDEX, i);
		outportb(VGA_GC_DATA, *regs);
		regs++;
	}
/* write ATTRIBUTE CONTROLLER regs */
	for(i = 0; i < VGA_NUM_AC_BYTES; i++)
	{
		(void)inportb(VGA_INSTAT_READ);
/* b5=0 here allows writing the first 16 AC registers
(the 16-color palette) but blanks the display */
		outportb(VGA_AC_INDEX, i);
		outportb(VGA_AC_WRITE, *regs);
		regs++;
	}
/* lock 16-color palette and unblank display */
	(void)inportb(VGA_INSTAT_READ);
	outportb(VGA_AC_INDEX, 0x20);
/* write DAC */
	outportb(VGA_DAC_WRITE_INDEX, 0);
	for(i = 0; i < VGA_NUM_DAC_BYTES; i += 3)
	{
/* VGA_DAC_WRITE_INDEX auto-increments after writing
three bytes to VGA_DAC_DATA */
		outportb(VGA_DAC_DATA, regs[0]);
		outportb(VGA_DAC_DATA, regs[1]);
		outportb(VGA_DAC_DATA, regs[2]);
		regs += 3;
	}
}
/****************************************************************************/
static void read_regs(unsigned char *regs)
{
	unsigned i;

/* read MISCELLANEOUS reg */
	*regs = inportb(VGA_MISC_READ);
	regs++;
/* read SEQUENCER regs */
	for(i = 0; i < VGA_NUM_SEQ_BYTES; i++)
	{
		outportb(VGA_SEQ_INDEX, i);
		*regs = inportb(VGA_SEQ_DATA);
		regs++;
	}
/* read CRTC regs */
	for(i = 0; i < VGA_NUM_CRTC_BYTES; i++)
	{
		outportb(VGA_CRTC_INDEX, i);
		*regs = inportb(VGA_CRTC_DATA);
		regs++;
	}
/* read GRAPHICS CONTROLLER regs */
	for(i = 0; i < VGA_NUM_GC_BYTES; i++)
	{
		outportb(VGA_GC_INDEX, i);
		*regs = inportb(VGA_GC_DATA);
		regs++;
	}
/* read ATTRIBUTE CONTROLLER regs */
	for(i = 0; i < VGA_NUM_AC_BYTES; i++)
	{
		(void)inportb(VGA_INSTAT_READ);
/* nope, AC registers must be locked for read as well as write
		outportb(VGA_AC_INDEX, i | 0x20); */
		outportb(VGA_AC_INDEX, i);
		*regs = inportb(VGA_AC_READ);
		regs++;
	}
/* lock 16-color palette and unblank display */
	(void)inportb(VGA_INSTAT_READ);
	outportb(VGA_AC_INDEX, 0x20);
/* read DAC */
	outportb(VGA_DAC_READ_INDEX, 0);
	for(i = 0; i < VGA_NUM_DAC_BYTES; i += 3)
	{
/* VGA_DAC_READ_INDEX auto-increments after reading
three bytes from VGA_DAC_DATA */
		regs[0] = inportb(VGA_DAC_DATA);
		regs[1] = inportb(VGA_DAC_DATA);
		regs[2] = inportb(VGA_DAC_DATA);
		regs += 3;
	}
}
/****************************************************************************/
static void set_plane(unsigned p)
{
	unsigned char pmask;

	p &= 3;
	pmask = 1 << p;
/* set read plane */
	outportb(VGA_GC_INDEX, 4);
	outportb(VGA_GC_DATA, p);
/* set write plane */
	outportb(VGA_SEQ_INDEX, 2);
	outportb(VGA_SEQ_DATA, pmask);
}
/*****************************************************************************
write font to VGA memory plane P4 (assuming planes are named P1, P2, P4, P8)
*****************************************************************************/
static void write_font(const unsigned char *buf, unsigned font_height)
{
	unsigned char seq2, seq4, gc4, gc5, gc6;
	unsigned i;

/* save registers modified by this routine and by set_plane() */
	outportb(VGA_GC_INDEX, 4);	gc4 = inportb(VGA_GC_DATA);
	outportb(VGA_GC_INDEX, 5);	gc5 = inportb(VGA_GC_DATA);
	outportb(VGA_GC_INDEX, 6);	gc6 = inportb(VGA_GC_DATA);
	outportb(VGA_SEQ_INDEX, 2);	seq2 = inportb(VGA_SEQ_DATA);
	outportb(VGA_SEQ_INDEX, 4);	seq4 = inportb(VGA_SEQ_DATA);
/* turn off even-odd addressing (set flat addressing)
assume: chain-4 addressing already off */
	outportb(VGA_SEQ_INDEX, 4);	outportb(VGA_SEQ_DATA, seq4 | 0x04);
	outportb(VGA_GC_INDEX, 5);	outportb(VGA_GC_DATA, gc5 & ~0x10);
	outportb(VGA_GC_INDEX, 6);	outportb(VGA_GC_DATA, gc6 & ~0x02);
/* write font to plane P4 */
	set_plane(2);
	for(i = 0; i < 256; i++)
	{
		memcpy((char *)0xB8000L + 16384u * /*font num.=*/0
			+ i * 32, buf, font_height);
		buf += font_height;
	}
/* restore registers */
	outportb(VGA_GC_INDEX, 4);	outportb(VGA_GC_DATA, gc4);
	outportb(VGA_GC_INDEX, 5);	outportb(VGA_GC_DATA, gc5);
	outportb(VGA_GC_INDEX, 6);	outportb(VGA_GC_DATA, gc6);
	outportb(VGA_SEQ_INDEX, 2);	outportb(VGA_SEQ_DATA, seq2);
	outportb(VGA_SEQ_INDEX, 4);	outportb(VGA_SEQ_DATA, seq4);
}
/*****************************************************************************
read font from VGA memory plane P4
*****************************************************************************/
static void read_font(unsigned char *buf, unsigned font_height)
{
	unsigned char seq2, seq4, gc4, gc5, gc6;
	unsigned i;

/* save registers modified by this routine and by set_plane() */
	outportb(VGA_GC_INDEX, 4);	gc4 = inportb(VGA_GC_DATA);
	outportb(VGA_GC_INDEX, 5);	gc5 = inportb(VGA_GC_DATA);
	outportb(VGA_GC_INDEX, 6);	gc6 = inportb(VGA_GC_DATA);
	outportb(VGA_SEQ_INDEX, 2);	seq2 = inportb(VGA_SEQ_DATA);
	outportb(VGA_SEQ_INDEX, 4);	seq4 = inportb(VGA_SEQ_DATA);
/* turn off even-odd addressing (set flat addressing)
assume: chain-4 addressing already off */
	outportb(VGA_SEQ_INDEX, 4);	outportb(VGA_SEQ_DATA, seq4 | 0x04);
	outportb(VGA_GC_INDEX, 5);	outportb(VGA_GC_DATA, gc5 & ~0x10);
	outportb(VGA_GC_INDEX, 6);	outportb(VGA_GC_DATA, gc6 & ~0x02);
/* read font from plane P4 */
	set_plane(2);
	for(i = 0; i < 256; i++)
	{
		memcpy(buf, (char *)0xB8000L + 16384u * /*font num.=*/0
			+ i * 32, font_height);
		buf += font_height;
	}
/* restore registers */
	outportb(VGA_GC_INDEX, 4);	outportb(VGA_GC_DATA, gc4);
	outportb(VGA_GC_INDEX, 5);	outportb(VGA_GC_DATA, gc5);
	outportb(VGA_GC_INDEX, 6);	outportb(VGA_GC_DATA, gc6);
	outportb(VGA_SEQ_INDEX, 2);	outportb(VGA_SEQ_DATA, seq2);
	outportb(VGA_SEQ_INDEX, 4);	outportb(VGA_SEQ_DATA, seq4);
}
/****************************************************************************/
typedef struct
{
	unsigned wd, ht;
	unsigned long bytes_per_row;
	unsigned char *raster;
} img_t;

static img_t g_fb;

void set_graphics_mode(void)
{
/* VGA video mode 0x13 */
	static unsigned char _320x200x256[VGA_NUM_BYTES] =
	{
/* MISC */
		0x63,
/* SEQ */
		0x03, 0x01, 0x0F, 0x00, 0x0E,
/* CRTC */
		0x5F, 0x4F, 0x50, 0x82, 0x54, 0x80, 0xBF, 0x1F,
		0x00, 0x41, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x9C, 0x0E, 0x8F, 0x28,	0x40, 0x96, 0xB9, 0xA3,
		0xFF,
/* GC */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F,
		0xFF,
/* AC */
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
		0x41, 0x00, 0x0F, 0x00,	0x00
/* DAC
		...values here don't matter, will be written from BMP file. */
	};

/* save VGA registers (video mode, 25 or 50 rows, and hardware cursor pos'n)
and font */
	read_regs(g_text);
/*	read_font(g_font, 8);
	read_font(g_font, 16); */
	read_font(g_font, (g_vc_ht == 50) ? 8 : 16);
	g_text_init = 1;

	write_regs(_320x200x256);
	g_fb.wd = 320;
	g_fb.ht = 200;
	g_fb.bytes_per_row = 320;
	g_fb.raster = (unsigned char *)0xA0000L;
}
/****************************************************************************/
void set_text_mode(void)
{
	if(!g_text_init)
		return;
	write_regs(g_text);
	write_font(g_font, (g_vc_ht == 50) ? 8 : 16);
	clear_screen();
}
/****************************************************************************/
static unsigned read_le16(void *buf_ptr)
{
	unsigned char *buf = buf_ptr;
	unsigned rv;

	rv = buf[1]; /* MSB */
	rv <<= 8;
	rv |= buf[0]; /* LSB */
	return rv;
}
/****************************************************************************/
static unsigned long read_le32(void *buf_ptr)
{
	unsigned char *buf = buf_ptr;
	unsigned long rv;

	rv = buf[3]; /* MSB */
	rv <<= 8;
	rv |= buf[2];
	rv <<= 8;
	rv |= buf[1];
	rv <<= 8;
	rv |= buf[0]; /* LSB */
	return rv;
}
/****************************************************************************/
void display_bmp_image(unsigned char *buf)
{
	unsigned i, xoff, yoff, y;
	unsigned long pal_off;
	img_t img;

// xxx - width & height are actually 32-bit values in the .BMP file:
	img.wd = read_le16(&buf[18]);
	img.ht = read_le16(&buf[22]);
	img.bytes_per_row = (img.wd + 3) & ~3L;
/* 32-bit raster offset in .BMP header: */
	img.raster = &buf[read_le32(&buf[10])];
/* dword at offset 14 is actually header size -- palette follows header */
	pal_off = 14 + read_le32(&buf[14]);
/* set 256-color palette */
#define	VGA_DAC_WRITE_INDEX	0x3C8
#define	VGA_DAC_DATA		0x3C9
	outportb(VGA_DAC_WRITE_INDEX, 0);
	for(i = 0; i < 256; i++)
	{
		outportb(VGA_DAC_DATA, buf[pal_off + i * 4 + 2] / 4);
		outportb(VGA_DAC_DATA, buf[pal_off + i * 4 + 1] / 4);
		outportb(VGA_DAC_DATA, buf[pal_off + i * 4 + 0] / 4);
	}
#define min(a,b)    	(((a) < (b)) ? (a) : (b))
	xoff = yoff = 0;
/* blit */
	while(1)
	{
		for(y = 0; y < min(img.ht, g_fb.ht); y++)
		{
			memcpy(&g_fb.raster[g_fb.bytes_per_row * y],
/* .BMP raster is upside-down: */
				&img.raster[img.bytes_per_row
					* (img.ht - (y + yoff) - 1) + xoff],
				min(img.bytes_per_row, g_fb.bytes_per_row));
		}
/* get key */
		i = getch();
/* raw scancodes; not BIOS-translated!
		if(i == 0) */
		if(i == 0xE0)
			i = 0x100 | getch();
/* Esc quits *
		if(i == 27) */
		if(i == 1)
			break;
#define	KEY_HOME	0x147
#define	KEY_UPARROW	0x148
#define	KEY_PAGEUP	0x149
#define KEY_LEFTARROW	0x14B
#define KEY_RIGHTARROW	0x14D
#define	KEY_END		0x14F
#define	KEY_DOWNARROW	0x150
#define	KEY_PAGEDOWN	0x151
#define	DELTA_X		32
#define	DELTA_Y		32
/* horizontal (left & right) scrolling */
		if(img.wd > g_fb.wd)
		{
			if(i == KEY_RIGHTARROW)
			{
				if(xoff + DELTA_X > img.wd - g_fb.wd)
					xoff = img.wd - g_fb.wd;
				else
					xoff += DELTA_X;
			}
			else if(i == KEY_LEFTARROW)
			{
				if(xoff < DELTA_X)
					xoff = 0;
				else
					xoff -= DELTA_X;
			}
			else if(i == KEY_HOME)
				xoff = 0;
			else if(i == KEY_END)
				xoff = img.wd - g_fb.wd;
		}
/* vertical (up & down) scrolling */
		if(img.ht > g_fb.ht)
		{
			if(i == KEY_DOWNARROW)
			{
				if(yoff + DELTA_Y > img.ht - g_fb.ht)
					yoff = img.ht - g_fb.ht;
				else
					yoff += DELTA_Y;
			}
			else if(i == KEY_UPARROW)
			{
				if(yoff < DELTA_Y)
					yoff = 0;
				else
					yoff -= DELTA_Y;
			}
			else if(i == KEY_PAGEUP)
				yoff = 0;
			else if(i == KEY_PAGEDOWN)
				yoff = img.ht - g_fb.ht;
		}
	}
}
/*----------------------------------------------------------------------------
KRNL/MAIN.C
------------------------------------------------------------------------------
#include <hardware.h> inportb(), outportb()
#include <stdio.h> printf() */

/* IMPORTS
from KSTART.S */
extern mboot_info_t *g_mboot_info;

/* from CONSOLE.C */
unsigned getch(void);

/* from VGA.C */
void set_graphics_mode(void);
void set_text_mode(void);
void display_bmp_image(unsigned char *image);
/*****************************************************************************
On my Pentium system, A20 must be enabled for reboot() to work.
When I used this function in a real-mode kernel, it gave me
exception 6 (invalid instruction) instead of rebooting.
*****************************************************************************/
static void reboot(void)
{
	unsigned i;

//	disable();
/* flush the keyboard controller */
	do
	{
		i = inportb(0x64);
		if(i & 0x01)
		{
			(void)inportb(0x60);
			continue;
		}
	} while(i & 0x02);
/* pulse the CPU reset line */
	outportb(0x64, 0xFE);
}
/****************************************************************************/
#define BPERL		16	/* byte/line for dump */

static void dump(void *data_p, unsigned count)
{
	unsigned char *data = (unsigned char *)data_p;
	unsigned byte1, byte2;

	while(count != 0)
	{
		for(byte1 = 0; byte1 < BPERL; byte1++)
		{
			if(count == 0)
				break;
			printf("%02X ", data[byte1]);
			count--;
		}
		printf("\t");
		for(byte2 = 0; byte2 < byte1; byte2++)
		{
			if(data[byte2] < ' ')
				printf(".");
			else
				printf("%c", data[byte2]);
		}
		printf("\n");
		data += BPERL;
	}
}
/****************************************************************************/
void kmain(void)
{
	unsigned char *buf;
/*	unsigned long len;*/
	mboot_mod_t *m;
	unsigned i;

	init_video();
	printf("MB-TEST by Chris Giese <geezer@execpc.com> "
		"http://my.execpc.com/~geezer\nreleased Feb. 25, 2014\n");
/* memory fields */
	if(g_mboot_info->flags & MBINFO_MEM)
		printf("%luK conventional memory, %luK extended memory,\n",
			g_mboot_info->conv_mem, g_mboot_info->ext_mem);
	else
		printf("*** Bootloader did not set memory sizes\n");
/* root device */
	if(g_mboot_info->flags & MBINFO_ROOT_DEV)
		printf("Root device = 0x%lX\n", g_mboot_info->root_dev);
	else
		printf("*** Bootloader did not set root device\n");
/* kernel command line */
	if(g_mboot_info->flags & MBINFO_CMD_LINE)
		printf("Kernel command line (%u bytes at 0x%p):\n\t'%s'\n",
			strlen((char *)g_mboot_info->cmd_line) + 1,
			(char *)g_mboot_info->cmd_line,
			(char *)g_mboot_info->cmd_line);
	else
		printf("*** Bootloader did not set kernel command line\n");
/* modules */
	if(g_mboot_info->flags & MBINFO_MODS)
	{
		printf("%lu module(s) (%lu bytes at 0x%lX):\n"
			"\taddress  size     command line\n"
			"\t-------- -------- ----------------------\n",
			g_mboot_info->num_mods, g_mboot_info->num_mods
			* sizeof(mboot_mod_t), g_mboot_info->mods_addr);
		for(i = 0; i < g_mboot_info->num_mods; i++)
		{
			m = &((mboot_mod_t *)g_mboot_info->mods_addr)[i];
			printf("\t     %8lX %8lX '%s'\n", m->start_addr,
				m->end_addr - m->start_addr,
				(char *)m->cmd_line);
		}
	}
	else
		printf("*** Bootloader did not set module fields\n");
/* BIOS memory ranges from INT 15h AX=E820h */
	if(g_mboot_info->flags & MBINFO_MEM_MAP)
	{
		mboot_range_t *r;

		printf("Memory map (%lu bytes at 0x%lX):\n"
			"\tstart    length   end      type\n"
			"\t-------- -------- -------- ----\n",
			g_mboot_info->map_len, g_mboot_info->map_addr);
		for(i = 0; i < g_mboot_info->map_len; )
		{
			r = (mboot_range_t *)(g_mboot_info->map_addr + i);
			printf("\t%8lX %8lX %8lX %4lX\n", r->addr, r->len,
				r->addr + r->len - 1, r->type);
			i += (4 + r->size);
		}
	}
	else
		printf("*** Bootloader did not set memory map fields\n");
/* drive info */
	if(g_mboot_info->flags & MBINFO_DRIVE_INFO)
	{
		mboot_drive_t *d;

		printf(	"Hard drive info (%lu bytes at 0x%lX)\n"
			"\tnum   lba cyls  heads sects\n"
			"\t----- --- ----- ----- -----\n",
			g_mboot_info->drives_len, g_mboot_info->drives_addr);
		for(i = 0; i < g_mboot_info->drives_len; )
		{
			d = (mboot_drive_t *)(g_mboot_info->drives_addr + i);
			printf("\t 0x%02X %3u %5u %5u %5u\n",
				d->int13_drive_num, d->use_lba, d->cyls,
				d->heads, d->sects);
			i += d->len;
		}
	}
	else
		printf("*** Bootloader did not set drive info fields\n");
/* ROM config table from INT 15h AH=C0h */
	if(g_mboot_info->flags & MBINFO_CONFIG_TABLE)
	{
		printf("ROM config table at 0x%lX\n",
			g_mboot_info->config_table);
	}
	else
		printf("*** Bootloader did not set ROM config table field\n");
/* loader name */
	if(g_mboot_info->flags & MBINFO_LOADER_NAME)
	{
		printf("Bootloader name (%u bytes at 0x%lX):\n\t'%s'\n",
			strlen((char *)g_mboot_info->loader_name) + 1,
			g_mboot_info->loader_name,
			(char *)g_mboot_info->loader_name);
	}
	else
		printf("*** Bootloader did not set bootloader name field\n");
	if(g_mboot_info->num_mods != 0)
		printf("While each image is displayed on the screen,\n"
			"use the arrow keys to scroll or press Esc to quit\n");
	else
		printf("No modules! Configure bootloader to load 256-color "
			".BMP files\nas Multiboot modules and this program "
			"will display them\n");
/* display images */
	for(i = 0; i < g_mboot_info->num_mods; i++)
	{
		m = &((mboot_mod_t *)g_mboot_info->mods_addr)[i];
/*		len = m->end_addr - m->start_addr + 1; */
		buf = (unsigned char *)m->start_addr;
/* validate BMP file */
		if(buf[0] != 'B' || buf[1] != 'M')
NOT:		{
			printf("Module %u is not a 256-color uncompressed "
				" BMP file\n", i + 1);
			printf("Dump of first few bytes in module:\n");
			dump(buf, 32);
			continue;
		}
		if(read_le16(&buf[26]) != 1	/* number of planes */
			|| read_le16(&buf[28]) != 8	/* bits per pixel */
			|| read_le32(&buf[30]) != 0)	/* compression */
				goto NOT;
		printf("Press a key to display image %u of %lu\n",
			i + 1, g_mboot_info->num_mods);
		getch();
		set_graphics_mode();
		display_bmp_image(buf);
		set_text_mode();
	}
	printf("All done. Press a key to reboot\n");
	getch();
	reboot();
	return;
}
